/*
 * =======================================================================
 * ============================ JAR WIZARD ===============================
 * =======================================================================
 *
 *                                                 Copyright (C) 2002 Chir
 *
 * DESCRIPTION:   This Swing program helps packing a Java application
 *                into a JAR file that then launches with 2 clicks.
 *
 * THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
 * IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
 * THE FREE SOFTWARE FOUNDATION; EITHER VERSION 2 OF THE LICENSE, OR
 * (AT YOUR OPTION) ANY LATER VERSION.
 *
 * THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL,
 * BUT WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  SEE THE
 * GNU GENERAL PUBLIC LICENSE FOR MORE DETAILS.
 *
 * YOU SHOULD HAVE RECEIVED A COPY OF THE GNU GENERAL PUBLIC LICENSE
 * ALONG WITH THIS PROGRAM; IF NOT, WRITE TO THE FREE SOFTWARE
 * FOUNDATION, INC., 59 TEMPLE PLACE, SUITE 330, BOSTON, MA 02111-1307 USA
 */

import java.io.*;
import java.util.*;
import org.chir.*;
import java.util.jar.*;

/**
 * Handles the creation of a JAR file.
 * It generates a temporary MANIFEST file if needed, and invokes the JAR command.
 *
 * @author  Chir
 */
public class JarCreator
{
    /**
     * Exception subclass thrown by the createJar function.
     */
    public class JarCreatorException extends Exception
    {
        public JarCreatorException( String s )  { super(s); }
    }

    /**
     * Declares a file extension to be included in the final JAR file.
     * All files with the given extension will later be included when generating the JAR file.
     * @param extName       The extension to look for (not including the dot separator).
     */
    public void addExtension( String extName )
    {
        if( extName != null  &&  extName.trim().length() > 0 )  m_extNames.add( "."+extName.trim().toLowerCase() );
    }

    /**
     * Declares a specific file to be included in the final JAR file.
     * @param fileName       The file name (not including the path).
     */
    public void addFile( String fileName )
    {
        assert( fileName != null  &&  fileName.trim().length() > 0 );
        m_fileNames.add( fileName.trim() );
    }

    /**
     * Launches the execution of the JAR command, using the specified filenames.
     * @param dirName   The path where all the classes are located (root classpath of the Application to pack).
     * @param jarName   The path to the new JAR file to create.
     * @param mainName  The path to the main class (for application or beans) or NULL if none (for applets or libraries).
     * @param isBean    True if packing a Bean instead of an Application.
     * @return          The full command-line, equivalent to the operation performed.
     */
    public String createJar( String dirName, String jarName, String mainName, boolean isBean, String classpathName, int zipLevel )  throws JarCreatorException
    {
        //System.out.println( "\ncreateJar() jarName=|" + jarName + "|\n" );
/*
        if( ! checkCommand( "jar --help" ) )
        {
            throw new JarCreatorException( ResourceBundle.getBundle("jarwizard_res").getString("IDS_JAR_NOT_FOUND_ERROR") );
        }
 */
        if( ! jarName.toLowerCase().endsWith(".jar") )
        {
            jarName = jarName + ".jar";         // Ensure that the file name has the JAR extension.
        }
/*
        if( jarName.indexOf(' ') >= 0 )
        {
            jarName = "\"" + jarName + "\"";    // Enclose in quotes, only if spaces are found.
        }
*/
        if( dirName.charAt( dirName.length() - 1 ) == '/'  ||  dirName.charAt( dirName.length() - 1 ) == '\\' )
        {
            dirName = dirName.substring( 0, dirName.length() - 1 ); // Remove trailing slash.
        }

        if( mainName != null  &&  mainName.equals("") )  mainName = null;

        if( mainName != null )
        {
            if( mainName.endsWith(".class") )
            {
                mainName = mainName.substring( 0, mainName.lastIndexOf(".class") ); // Remove the CLASS extension from the name.
            }

            if( mainName.startsWith(dirName) )
            {
                mainName = mainName.substring( dirName.length() + 1 );  // Remove current path from the name.
            }

            mainName = mainName.replace('\\','/');  // Replace all MSDOS slashes with unix/mac slashes.
        }

        File dirF = null;           // Base directory.
        OutputStream outF = null;   // manifest.txt

        if( dirName != null )
        {
            dirF = new File( dirName );
            if( ! dirF.exists() )  throw new JarCreatorException( ResourceBundle.getBundle("jarwizard_res").getString("IDS_DIR_NOT_FOUND_ERROR") );
        }

        // --- Create the manifest file ---
        if( mainName != null )
        {
            try
            {
                String t = "";

                t += "Manifest-Version: 1.0\n";

                if( isBean )
                {
                    t += "Name: " + mainName + ".class\nJava-Bean: True\n\n";
                    t += "Name: " + mainName + ".class\nJava-Bean: True\n\n";   // Needed for Forte/Netbeans compatibility :-(
                }
                else
                {
                    t += "Main-Class: " + mainName + "\n";
                }
                if( classpathName != null  &&  ! classpathName.trim().equals("") )
                {
                    t += "Class-Path: " + classpathName + "\n";
                }
                t += "\n";
                outF = new FileOutputStream( dirName + "/manifest.txt" );
                outF.write( t.getBytes() );
                outF.close();
            }
            catch( IOException e )
            {
                try { if(outF!=null) outF.close(); } catch(IOException e2)  { /* do nothing. */ }
                throw new JarCreatorException( ResourceBundle.getBundle("jarwizard_res").getString("IDS_WRITE_MANIFEST_ERROR") );  // "Can't write manifest.txt" );
            }
        }

        // --- Create the JAR file ---
        JarOutputStream zip_out;
        String cmdT;
        try
        {
            FileOutputStream out = new FileOutputStream( jarName );
            if( mainName != null )
            {
                Manifest manif = new Manifest( new FileInputStream( dirName + "/manifest.txt" ) );
                zip_out = new JarOutputStream( new BufferedOutputStream(out), manif );
                cmdT = "jar -cmf manifest.txt " + jarName + " ";
            }
            else
            {
                zip_out = new JarOutputStream( new BufferedOutputStream(out) );
                cmdT = "jar -cf " + jarName + " ";
            }
            zip_out.setLevel( zipLevel );
            cmdT += generateCommandString( dirF, true, zip_out );
            zip_out.close();
        }
        catch( IOException e )
        {
            throw new JarCreatorException( ResourceBundle.getBundle("jarwizard_res").getString("IDS_CREATE_JAR_ERROR") );
        }
        return cmdT;
/*
        // --- Create Jar file using JAR command ---
        String cmdT;
        if( mainName != null )
        {
            cmdT = "jar -cmf manifest.txt " + jarName + " " + generateCommandString( dirF, true, zip_out );
        }
        else
        {
            cmdT = "jar -cf " + jarName + " " + generateCommandString( dirF, true, zip_out );
        }

        System.out.println( "\nexec |" + cmdT + "| [" + dirF.toString() + "]\n" );
        try
        {
            if( ! CommandLine.ExecuteCommand( cmdT, dirF, 2, false, true ) )
            {
                throw new JarCreatorException( ResourceBundle.getBundle("jarwizard_res").getString("IDS_CREATE_JAR_ERROR") );
            }
            else
            {
                // --- Delete Manifest file ---
                // ...
            }
            zip_out.close();
        }
        catch( java.io.IOException e1 )
        {
            throw new JarCreatorException( ResourceBundle.getBundle("jarwizard_res").getString("IDS_JAR_NOT_FOUND_ERROR") );
        }
 */
    }

    private String getRelativePath( File dirF, File baseDirF )
    {
        try
        {
            String dirName = dirF.getCanonicalPath(); //.replace('\\','/');

            if( dirName.startsWith( baseDirF.getCanonicalPath() ) )
            {
                char c = baseDirF.getCanonicalPath().charAt( baseDirF.getCanonicalPath().length() - 1 );

                if( c == '/'  ||  c == '\\' )
                {
                    dirName = dirName.substring( baseDirF.getCanonicalPath().length() );
                }
                else
                {
                    if( dirName.length() > baseDirF.getCanonicalPath().length() )
                    {
                        dirName = dirName.substring( baseDirF.getCanonicalPath().length() + 1 );
                    }
                    else
                    {
                        dirName = "";
                    }
                }
            }
            if( ! dirName.equals("") )  { dirName += "/"; }
            return dirName;
        }
        catch( IOException e )
        {
            return dirF.getAbsolutePath() + "/";
        }
    }

    private void putNextJarEntry( JarOutputStream zip_out, File f, String zipName )
    {
        JarEntry zip_e2 = new JarEntry( zipName.replace('\\','/') );
        try
        {
            zip_out.putNextEntry( zip_e2 );
            if( ! f.isDirectory() )
            {
                BufferedInputStream b_in = new BufferedInputStream( new FileInputStream(f.getAbsolutePath()) );
                final int BUFFER_SIZE = 30000;
                byte[] data = new byte[BUFFER_SIZE];
                int count;
                while( (count=b_in.read(data, 0, BUFFER_SIZE)) != -1 )
                {
                    zip_out.write(data, 0, count);
                }
                b_in.close();
            }
            zip_out.closeEntry();
        }
        catch( IOException e )
        {
            System.out.println( "putNextEntry IOException " + e.toString() );
        }
    }

    private String generateCommandString( File f0, File baseDirF, boolean useWildcard, JarOutputStream zip_out )
    {
        File[]      	fileList = f0.listFiles();
		HashSet<String>	hasExt = new HashSet<String>();
        boolean     	hasSpace = ( (getRelativePath(f0,baseDirF) ).indexOf(' ') >= 0 );
        String      	s = "";

		for( File f : fileList )
        {
            if( f.isDirectory() )
            {
                if( ! f.getName().equals("CVS") ) // Only consider non-CVS directories.
                {
                    s += generateCommandString( f, baseDirF, useWildcard, zip_out );
                }
            }
            else
            {
                // --- Add this filename, if it is part of the wildcard list ---
                boolean found = false;
                for( String ext : m_extNames )
                {
                    if( f.getAbsolutePath().toLowerCase().endsWith(ext) )
                    {
                        if( hasSpace )
                        {
                            // Enclose this individual filename in quotes.
                            s += " \"" + getRelativePath(f0,baseDirF) + f.getName() + "\" ";
                        }
                        else
                        {
                            if( ! useWildcard )
                            {
                                // Add this individual filename.
                                s += " " + getRelativePath(f0,baseDirF) + f.getName() + " ";
                            }
                            else
                            {
                                // Use a wildcard instead, to make the final command-line simpler.
                                if( ! hasExt.contains(ext) )
                                {
                                	hasExt.add( ext );	// Don't add it again at this level.
                                    s += " " + getRelativePath(f0,baseDirF) + "*" + ext + " ";
                                }
                            }
                        }
                        putNextJarEntry( zip_out, f, getRelativePath(f0,baseDirF) + f.getName() );
                        found = true;
                        break;
                    }
                }
                if( ! found )
                {
                    // --- Add this individual filename, if it is part of the special list ---
                    if( m_fileNames.contains(f.getName()) )
                    {
                        if( hasSpace )
                        {
                            // Enclose this individual filename in quotes.
                            s += " \"" + getRelativePath(f0,baseDirF) + f.getName() + "\" ";
                        }
                        else
                        {
                            // No quotes are needed, so avoid them.
                            s += " " + getRelativePath(f0,baseDirF) + f.getName() + " ";
                        }
                        putNextJarEntry( zip_out, f, getRelativePath(f0,baseDirF) + f.getName() );
                    }
                }
            }
        }
        return s;
    }

    private String generateCommandString( File f, boolean useWildcard, JarOutputStream zip_out )
    {
        return generateCommandString( f, f, useWildcard, zip_out );
    }
 
    // -----------------------------------------------------------------------
    private Vector<String>	m_extNames = new Vector<String>();		// Vector<String> for all file types to include in the final JAR.
    private HashSet<String> m_fileNames = new HashSet<String>();    // Set<String> for all specific files to include in the final JAR.
}
